import sys
from .driver import Driver
from .util import import_symbol


def import_symbol_maybe(ob_or_path, sep=":"):
    if not isinstance(ob_or_path, str):
        return ob_or_path
    return import_symbol(ob_or_path, sep=sep)


def as_command(fn=None, argv=None, driver=Driver, level=2, _force=False):
    create_driver = import_symbol_maybe(driver)
    if argv is None:
        argv = sys.argv[1:]

    def call(fn, level=1, argv=argv):
        if not _force:
            # caller module extraction, if it is __main__, calling as command
            frame = sys._getframe(level)
            name = frame.f_globals["__name__"]
            if name != "__main__":
                return fn
        driver = create_driver(fn)
        return driver.run(argv)

    if fn is None:
        return call
    else:
        return call(fn, level=level, argv=argv)


# alias
handofcats = as_command



import sys
import re
import inspect
import functools
from prestring.python import (
    Module,
    LazyArgumentsAndKeywords,
)
from ._callback import CallbackArgumentParser


class _UnRepr:
    def __init__(self, val):
        self.val = val

    def __repr__(self):
        return str(self.val)

    __str__ = __repr__


def print_argparse_code(fn, history):
    def _make_args(history, default=""):
        name = history["name"]
        if name == "__init__":
            name = default
        kwargs = {k: (repr(v) if k != "type" else v.__name__) for k, v in history["kwargs"].items()}
        args = [repr(v) for v in history["args"]]
        return f"{name}({LazyArgumentsAndKeywords(args, kwargs)})"

    # xxx: is inplace?
    if "inplace" in history[0]["kwargs"]:
        inplace = history[0]["kwargs"].pop("inplace")

    m = Module()
    with m.def_("main", "argv=None"):
        m.import_("argparse")
        m.stmt(f"parser = argparse.ArgumentParser{_make_args(history[0])}")
        m.stmt("parser.print_usage = parser.print_help")
        for x in history[1:-1]:
            m.stmt(f"parser.{_make_args(x)}")

        history[-1] = {
            "name": history[-1]["name"],
            "args": (_UnRepr("argv"), ),
            "kwargs": {},
        }  # xxx

        m.stmt(f"args = parser.{_make_args(history[-1])}")
        m.stmt(f"{fn.__name__}(**vars(args))")

    with m.if_("__name__ == '__main__'"):
        m.stmt("main()")

    target_file = inspect.getsourcefile(fn)
    with open(target_file) as rf:
        source = rf.read()
    rx = re.compile("(?:^@([\S]+\.)?as_command.*|^.*import as_command.*)\n", re.MULTILINE)
    exposed = rx.sub("", "".join(source))

    def _dump(out):
        print(exposed, file=out)
        print(m, file=out)

    if not inplace:
        _dump(sys.stdout)
        sys.exit(0)
    else:
        import tempfile
        import os.path
        outpath = None
        try:
            with tempfile.NamedTemporaryFile("w", dir=".", delete=False) as wf:
                outpath = wf.name
                _dump(wf)
            dirpath = os.path.dirname(target_file) or "."
            os.makedirs(dirpath, exist_ok=True)
            os.rename(outpath, target_file)
            sys.exit(0)
        except Exception as e:
            print("error is occured. rollback (exception={e})".format(e=e))
            with open(target_file, "w") as wf:
                wf.write(source)
        finally:
            if os.path.exists(outpath):
                os.remove(outpath)
        sys.exit(1)


create_parser = functools.partial(CallbackArgumentParser, print_argparse_code)



class CallbackArgumentParser:
    def __init__(self, callback, fn, *args, history=None, **kwargs):
        self.callback = callback
        self.fn = fn
        self.history = history or [{"name": "__init__", "args": args, "kwargs": kwargs}]

    def __getattr__(self, name):
        self.history.append({"name": name})
        return self

    def __call__(self, *args, **kwargs):
        latest = self.history[-1]
        assert "args" not in latest
        latest["args"] = args
        latest["kwargs"] = kwargs

    def parse_args(self, *args, **kwargs):
        self.history.append({"name": "parse_args", "args": args, "kwargs": kwargs})
        self.callback(fn=self.fn, history=self.history)



import typing as t
import itertools
from logging import getLogger as get_logger
from .util import reify
logger = get_logger(__name__)


class Driver:
    def __init__(self, fn):
        self.fn = fn

    @reify
    def accessor(self):
        from .accessor import Accessor
        return Accessor(self.fn)

    def create_parser(self, fn, *, argv=None, description=None):
        if "--expose" in (argv or []):
            from .parsers import expose
            inplace = "--inplace" in (argv or [])
            description = description or fn.__doc__
            parser = expose.create_parser(fn, description=description, inplace=inplace)
        else:
            from .parsers import commandline
            parser = commandline.create_parser(fn, description=description or fn.__doc__)
            parser.add_argument("--expose", action="store_true")  # xxx (for ./expose.py)
            parser.add_argument("--inplace", action="store_true")  # xxx (for ./expose.py)
        return parser

    def _setup_type(self, opt, kwargs):
        if opt.type == bool:
            action = "store_true"
            if opt.default is True:
                action = "store_false"
            kwargs.pop("required", None)
            kwargs.pop("default", None)
            kwargs["action"] = action
        elif opt.type in (int, float):
            kwargs["type"] = opt.type
        else:
            from collections.abc import Sequence
            if hasattr(opt.type, "__origin__") and hasattr(opt.type, "__args__"):
                try:
                    # for Optional
                    nonetype = type(None)
                    if opt.type.__origin__ == t.Union and nonetype in opt.type.__args__ and len(
                        opt.type.__args__
                    ) == 2:
                        item_type = opt._replace(
                            type=[t for t in opt.type.__args__ if t is not nonetype][0]
                        )
                        self._setup_type(item_type, kwargs)

                    # for sequence (e.g. t.List[int], t.Tuple[str])
                    elif issubclass(opt.type.__origin__, Sequence):
                        kwargs["action"] = "append"
                        item_type = opt._replace(type=opt.type.__args__[0])
                        self._setup_type(item_type, kwargs)
                except:
                    logger.info(
                        "unexpected generic type is found (type=%s)", opt.type, exc_info=True
                    )
            elif hasattr(opt.type, "__supertype__"):  # for NewType
                # choices support (tentative)
                if hasattr(opt.type, "choices"):
                    kwargs["choices"] = opt.type.choices
                origin_type = opt._replace(type=opt.type.__supertype__)
                self._setup_type(origin_type, kwargs)
            elif issubclass(opt.type, (list, tuple)):
                kwargs["action"] = "append"
            else:
                logger.info("unexpected type is found (type=%s)", opt.type)

    def setup_parser(self, parser):
        arguments = [(opt, None) for opt in self.accessor.arguments]
        flags = [(opt, opt.required) for opt in self.accessor.flags]

        for opt, required in itertools.chain(arguments, flags):
            kwargs = {}
            if required is not None:
                kwargs["required"] = required
            if opt.default is not None:
                kwargs["default"] = opt.default
            if opt.type and opt.type != str:
                self._setup_type(opt, kwargs)
            if kwargs.get("action") == "append" and not opt.option_name.startswith("-"):
                kwargs["nargs"] = "*"
                kwargs.pop("action")

            if "default" in kwargs:
                kwargs["help"] = "(default: {!r})".format(kwargs["default"])

            logger.debug("add_argument %s %r", opt.option_name, kwargs)
            parser.add_argument(opt.option_name, **kwargs)

    def run(self, argv=None):
        fn = self.fn
        parser = self.create_parser(fn, argv=argv, description=fn.__doc__)
        self.setup_parser(parser)
        args = parser.parse_args(argv)
        params = vars(args).copy()
        params.pop("expose", None)  # xxx: for ./parsers/expose.py
        params.pop("inplace", None)  # xxx: for ./parsers/expose.py
        return fn(**params)



import functools
from ._callback import CallbackArgumentParser


class ParseArgsCalled(Exception):
    def __init__(self, *, fn, history):
        self.fn = fn
        self.history = history


def on_parse_args(*, fn, history):
    raise ParseArgsCalled(fn=fn, history=history)


create_parser = functools.partial(CallbackArgumentParser, on_parse_args)



import argparse
from logging import getLogger as get_logger
from . import as_command
from . import util
logger = get_logger(__name__)


def import_symbol(path, sep=":", logger=logger):
    try:
        return util.import_symbol(path, sep=sep, logger=logger)
    except ValueError as e:
        logger.info(str(e), exc_info=True)
        import argparse  # xxx
        raise argparse.ArgumentTypeError("must be in 'module:attrs' format")


def main(argv=None):
    parser = argparse.ArgumentParser(add_help=False)
    parser.print_usage = parser.print_help

    parser.add_argument(
        "entry_point",
        type=import_symbol,
        help="target EntryPoint. (must be in 'module:attrs' format)",
    )
    parser.add_argument(
        "--driver",
        type=import_symbol,
        default="handofcats.driver:Driver",
        help="driver class, this is experimental (default: handofcats.driver:Driver)",
    )

    args, rest_argv = parser.parse_known_args(argv)
    as_command(args.entry_point, argv=rest_argv, level=3, driver=args.driver)


if __name__ == "__main__":
    main()



import sys
from logging import getLogger as get_logger
from importlib import (
    import_module,
    machinery,
)
logger = get_logger(__name__)


# stolen from pyramid
class reify(object):
    """cached property"""

    def __init__(self, wrapped):
        self.wrapped = wrapped
        try:
            self.__doc__ = wrapped.__doc__
        except:
            pass

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val


def _import_symbol_from_module(module, name, logger=logger):
    logger.debug("import module=%s, name=%s", module, name)
    module = import_module(module)
    return getattr(module, name)


def _import_symbol_from_filepath(path, name, module_id=None, logger=logger):
    logger.debug("import path=%s, name=%s", path, name)
    module_id = module_id or path.replace("/", "_")
    if module_id.endswith(".py"):
        module_id = module_id[:-3]
    module = machinery.SourceFileLoader(module_id, path).load_module()
    return getattr(module, name)


def import_symbol(path, sep=":", logger=logger):
    try:
        module, name = path.rsplit(sep, 1)
        return _import_symbol_from_module(module, name, logger=logger)
    except (ImportError, TypeError):
        sys.path.append(".")
        return _import_symbol_from_filepath(module, name, logger=logger)



import os
from setuptools import setup, find_packages
here = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(here, 'README.rst')) as f:
        README = f.read()
    with open(os.path.join(here, 'CHANGES.txt')) as f:
        CHANGES = f.read()
except IOError:
    README = CHANGES = ''

install_requires = [
    "prestring",
]

docs_extras = []

tests_require = []

testing_extras = tests_require + []

setup(
    name='handofcats',
    version='2.2.0',
    description='python function to command translator',
    long_description=README + '\n\n' + CHANGES,
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
    ],
    keywords='',
    author="podhmo",
    author_email="",
    url="https://github.com/podhmo/handofcats",
    packages=find_packages(exclude=["handofcats.tests", "examples"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'testing': testing_extras,
        'docs': docs_extras,
    },
    tests_require=tests_require,
    license="mit",
    test_suite="handofcats.tests",
    entry_points="""
[console_scripts]
handofcats=handofcats.__main__:main
"""
)
